using System;
using System.Collections.Generic;

namespace InterpolationTest
{
    public class ClientSim
    {
        public bool isExtrapolating = false;
        public double renderTime = 0.0;
        public double clientTime = 0.0;
        public bool firstSnapshot = true;
        public bool smoothTimerState = true;

        private double physicsAcc = 0.0;
        private double renderAcc = 0.0;

        public Vector2D localPos1 = new Vector2D(0, 0);
        public Vector2D localVel1 = Settings.defaultVelocity;
        public Vector2D renderPos1 = new Vector2D(0, 0);
        public Vector2D renderVel1 = Settings.defaultVelocity;
        public Vector2D localPos2 = new Vector2D(0, 0);
        public Vector2D localVel2 = Settings.defaultVelocity;
        public Vector2D renderPos2 = new Vector2D(0, 0);
        public Vector2D renderVel2 = Settings.defaultVelocity;

        public List<Snapshot> snapshots = new List<Snapshot>();
        public double[] measuredLatency = new double[Settings.sampleAmmount];
        public int measuredLatencyIndex;
        public double targetTime;

        public DateTime targetSetTime;
        private System.Diagnostics.Stopwatch swex = new System.Diagnostics.Stopwatch();
        private System.Diagnostics.Stopwatch swup = new System.Diagnostics.Stopwatch();

        private int findSnapshot(double t)
        {
            for (int i = 0; i < snapshots.Count - 1; ++i)
            {
                Snapshot snapshot0 = snapshots[i + 0];
                Snapshot snapshot1 = snapshots[i + 1];
                if (t >= snapshot0.m_time && t <= snapshot1.m_time)
                {
                    return i;
                }
            }

            return -1;
        }

        public void addSnapshot(double snapTime, Vector2D pos, Vector2D vel)
        {
            if (firstSnapshot)
            {
                firstSnapshot = false;

                renderTime = snapTime - Settings.timeOffset;
                targetTime = renderTime;
                localPos2 = pos;
                localVel2 = vel;
            }

            Snapshot snapshot = new Snapshot();
            snapshot.m_time = snapTime;
            snapshot.m_pos = pos;
            snapshot.m_vel = vel;
            snapshots.Add(snapshot);

            setTargetTime(snapTime);
        }

        public void setTargetTime(double snapTime)
        {
            //if (targetTime == renderTime)
            //{
                measuredLatency[measuredLatencyIndex] = clientTime - snapTime;

                if (measuredLatencyIndex == Settings.sampleAmmount - 1)
                {
                    measuredLatencyIndex = 0;
                }
                else
                {
                    ++measuredLatencyIndex;
                }

                double maxLatency = 0;
                double sumLatency = 0;
                for (int i = 0; i < Settings.sampleAmmount; ++i)
                {
                    maxLatency = Math.Max(maxLatency, Math.Abs(measuredLatency[i]));
                    sumLatency += measuredLatency[i];
                }

                double avrLatency = sumLatency / Settings.sampleAmmount;
                //System.Diagnostics.Debug.WriteLine("avrLatency: " + avrLatency);

                double differenceTime = Math.Abs(renderTime - targetTime);

                if (swex.Elapsed.TotalMilliseconds / 1000 >= Settings.exptapoTime || differenceTime > Settings.reactDifference)
                {
                    targetTime = clientTime - avrLatency - Settings.timeOffset;
                    //targetTime = snapTime  - Settings.timeOffset + maxDifferenceCorrected;
                    //targetTime = snapTime - Settings.timeOffset;

                    System.Diagnostics.Debug.WriteLine("setTargetTime: SET");
                }
            //}
        }

        // Lerp that is framerate-independent.See https://www.scirra.com/blog/ashley/17/using-lerp-with-delta-time for the explanation and maths !
        public double lerpDT(double a, double b, double f, double dt)
        {
            double t = 1.0 - Math.Pow(f, dt);
            return ((a) * (1 - (t)) + (b) * (t));
        }

        public void update(double elapsed)
        {
            //System.Diagnostics.Debug.WriteLine("elapsedTime: " + elapsed);

            clientTime += elapsed;
            renderTime += elapsed;
            targetTime += elapsed;


            // simulate physics
            physicsAcc += elapsed;
            while (physicsAcc >= Settings.physicsStepClient)
            {
                physicsAcc -= Settings.physicsStepClient;
                localPos1 = localPos1 + localVel1 * Settings.physicsStepClient; // move player 1 (that one is local only, no interpolation)
                //localPos2 = localPos2 + localVel2 * Settings.physicsStepClient; // move player 2 (that's the one server-interpolated)
            }

            if (smoothTimerState)
            {
                // speed up or slow down the display time to match the target time
                // that's pretty much the cause of the stuttering, however without that code, changes in latency mean the client and server timers are no longer synchronized properly
                // ie. try to PageUp to increase the latency to 1 second -> the client extrapolates all the time (green dot filled)
                double delta = Math.Abs(renderTime - targetTime) * 3.0;
                delta = Math.Max(delta, 0.0);
                delta = Math.Min(delta, 1.0);
                //System.Diagnostics.Debug.WriteLine("delta: " + delta);
                renderTime = lerpDT(renderTime, targetTime, delta, elapsed);
                //displayTime = targetTime;
            }

            // process local snapshots
            while (snapshots.Count > 0)
            {
                Snapshot first = snapshots[0];
                if (first.m_time < renderTime - Settings.snapStoreTime)
                {
                    // discard too old snapshots
                    snapshots.RemoveAt(0);
                }
                else
                {
                    break;
                }
            }


            // Render and Snapshot usage
            int currentSnapshotIDrender = findSnapshot(renderTime);
            if (currentSnapshotIDrender == -1)
            {
                // extrapolate latest snapshot
                isExtrapolating = true;

                swex.Start();

                renderPos1 = localPos1;

                if (snapshots.Count > 0)
                {
                    Snapshot lastSnapshot = snapshots[snapshots.Count - 1];

                    double t = renderTime - lastSnapshot.m_time;
                    renderPos2 = lastSnapshot.m_pos + lastSnapshot.m_vel * t;
                }
            }
            else
            {
                isExtrapolating = false;

                swex.Reset();

                renderPos1 = localPos1;

                //System.Diagnostics.Debug.WriteLine("snapshots[currentSnapshotIDrender].m_time: " + snapshots[currentSnapshotIDrender].m_time);

                Snapshot snapshot0 = snapshots[currentSnapshotIDrender + 0];
                Snapshot snapshot1 = snapshots[currentSnapshotIDrender + 1];

                /*
                // interpolate between current and next snapshot



                double tr = (renderTime - snapshot0.m_time) / (snapshot1.m_time - snapshot0.m_time);
                tr = Math.Max(tr, 0.0);
                tr = Math.Min(tr, 1.0);

                System.Diagnostics.Debug.WriteLine("renderTime: " + renderTime);
                System.Diagnostics.Debug.WriteLine("snapshot0.m_time: " + snapshot0.m_time);
                System.Diagnostics.Debug.WriteLine("snapshot1.m_time: " + snapshot1.m_time);
                System.Diagnostics.Debug.WriteLine("tr: " + tr);

                renderPos2 = snapshot0.m_pos + (snapshot1.m_pos - snapshot0.m_pos) * tr;
                renderVel2 = snapshot0.m_vel + (snapshot1.m_vel - snapshot0.m_vel) * tr;
                //renderPos2 = snapshot0.m_pos;
                //renderVel2 = snapshot0.m_vel;
                */


                //Interpolate towards newest snap. Disregard older snap.
                double delta = Math.Abs(renderTime - snapshot1.m_time);

                System.Diagnostics.Debug.WriteLine("delta: " + delta);

                double f = delta;
                double dt = elapsed;
                double tr = 1.0 - Math.Pow(f, dt);
                tr = Math.Max(tr, 0.0);
                tr = Math.Min(tr, 1.0);

                double tp = 1.0 - Math.Pow(f, tr);
                renderPos2 = ((renderPos2) * (1 - (tr)) + (snapshot1.m_pos) * (tr));
                //double tv = 1.0 - Math.Pow(f, Settings.renderStep);
                double tv = 1.0 - Math.Pow(f, tr);
                renderVel2 = ((renderPos2) * (1 - (tr)) + (snapshot1.m_vel) * (tr));

                //System.Diagnostics.Debug.WriteLine("renderPos2: " + renderPos2);
            }


        }
    }
}
