using System;
using System.Collections.Generic;
using System.Windows.Input;

namespace InterpolationTest
{
    public class ServerSim
    {
        // data for player 2 on server, note that we don't care about player 1, it only exists on client for viewing purposes
        public Vector2D serverPos2 = new Vector2D(0.5, 0);
        public Vector2D serverVel2 = Settings.defaultVelocity;
        private Random rnd = new Random();

        private double snapAcc = 0.0;
        private uint snapTickID = 0;
        public double currentLag = Settings.lag;
        private double physicsAcc = 0.0;

        public List<Snapshot> snapshots = new List<Snapshot>();

        public void update(double elapsed)
        {
            // process inputs
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.Up))
            {
                serverVel2.Y += elapsed;
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.Down))
            {
                serverVel2.Y -= elapsed;
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.Left))
            {
                serverVel2.X -= elapsed;
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.Right))
            {
                serverVel2.X += elapsed;
            }

            // simulate physics. Required to have a constant time step so that snapshots get sent to the server at a regular snapshot.m_time
            physicsAcc += elapsed;
            while (physicsAcc >= Settings.physicsStepServer)
            {
                physicsAcc -= Settings.physicsStepServer;
                serverPos2 = serverPos2 + serverVel2 * Settings.physicsStepServer; // move local player

                // simulates networking
                // generate snapshots to be sent to the client at a regular rate
                // snapshots have an artificial latency built in, decreasing over time. Client only processes them (to simulate the "receive") when they reach zero.
                snapAcc += Settings.physicsStepServer;
                if (snapAcc >= Settings.snapRate)
                {
                    snapAcc -= Settings.snapRate;

                    Snapshot snapshot = new Snapshot();
                    snapshot.m_time = (double)snapTickID * Settings.snapRate;
                    snapshot.m_pos = serverPos2;
                    snapshot.m_vel = serverVel2;
                    snapshot.m_latencyTimer = currentLag * (0.95 + 0.1026 * rnd.NextDouble()); // add a bit of random lag

                    snapshots.Add(snapshot);


                    //System.Diagnostics.Debug.WriteLine("snapshot Nr." + snapTickID + " m_time: " + snapshot.m_time);

                    snapTickID++;
                }

                // process snapshots latencies
                for (int i = 0; i < snapshots.Count; ++i)
                {
                    snapshots[i].m_latencyTimer = Math.Max(snapshots[i].m_latencyTimer - Settings.physicsStepServer, 0.0);
                }
            }

            if (Keyboard.IsKeyDown(System.Windows.Input.Key.PageUp))
            {
                currentLag = Math.Min(currentLag + elapsed, 1.0f);
            }
            if (Keyboard.IsKeyDown(System.Windows.Input.Key.PageDown))
            {
                currentLag = Math.Max(currentLag - elapsed, 0.0f);
            }
        }
    }
}
